#!/usr/bin/env perl

use strict;
use warnings;
use Getopt::Long;
use File::Basename;
use POSIX qw(strftime);
use List::MoreUtils qw(uniq);
use Scalar::Util qw(looks_like_number);
use Pod::Usage;
use Bio::Tools::GFF;
use IO::File;
use AGAT::Omniscient;

my $header = get_agat_header();
my $primaryTag=undef;
my $opt_output= undef;
my $opt_attribute = undef;
my $opt_test = undef;
my $opt_gff = undef;
my $opt_verbose = undef;
my $opt_help;

# OPTION MANAGMENT
my @copyARGV=@ARGV;
if ( !GetOptions( 'f|ref|reffile|gff=s' => \$opt_gff,
                  "p|type|l=s"          => \$primaryTag,
                  'a|att|attribute=s'   => \$opt_attribute,
                  'flip!'            => \$opt_test,
                  'o|output=s'          => \$opt_output,
                  'v|verbose!'          => \$opt_verbose,
                  'h|help!'             => \$opt_help ) )
{
    pod2usage( { -message => 'Failed to parse command line',
                 -verbose => 1,
                 -exitval => 1 } );
}

if ($opt_help) {
    pod2usage( { -verbose => 99,
                 -exitval => 0,
                 -message => "$header\n" } );
}

if ( ! $opt_gff or ! $opt_attribute ){
    pod2usage( {
           -message => "$header\nAt least 2 parameters are mandatory:\n1) Input reference gff file: -f\n".
           "2) An attribute tag: -a\n\n",
           -verbose => 0,
           -exitval => 2 } );
}

###############
# Manage Output

## FOR GFF FILE
my $gffout_ok; my $fhout_discarded ; my $ostreamReport;

if ($opt_output) {
  my ($outfile,$path,$ext) = fileparse($opt_output,qr/\.[^.]*/);

  # set file names
  my $outfile_ok = $path.$outfile.$ext;
  my $outfile_discarded = $path.$outfile."_discarded.txt";
  my $outfile_report = $path.$outfile."_report.txt";

  # check existence
  if(-f $outfile_ok){  print "File $outfile_ok already exist.\n";exit;}
  if(-f $outfile_discarded){  print "File $outfile_discarded already exist.\n";exit;}
  if(-f $outfile_report){  print "File $outfile_report already exist.\n";exit;}

  # create fh
  open( my $fh, '>', $outfile_ok) or die "Could not open file $outfile_ok $!";
  $gffout_ok = Bio::Tools::GFF->new(-fh => $fh, -gff_version => 3 );

  open( $fhout_discarded, '>', $outfile_discarded) or die "Could not open file $outfile_discarded $!";


  open($ostreamReport, '>', $outfile_report) or die "Could not open file $outfile_report $!";

}
else{
  $gffout_ok = Bio::Tools::GFF->new(-fh => \*STDOUT, -gff_version => 3);
  $fhout_discarded = Bio::Tools::GFF->new(-fh => \*STDOUT, -gff_version => 3);
  $ostreamReport = Bio::Tools::GFF->new(-fh => \*STDOUT, -gff_version => 3);
}

# Manage $primaryTag
my @ptagList;
my $print_feature_string;
if(! $primaryTag or $primaryTag eq "all"){
  $print_feature_string = "all features";
  push(@ptagList, "all");
}
elsif($primaryTag =~/^level[123]$/){
  $print_feature_string .= "$primaryTag features ";
  push(@ptagList, $primaryTag);
}
else{
   @ptagList= split(/,/, $primaryTag);
   foreach my $tag (@ptagList){
      if($tag =~/^level[123]$/){
        $print_feature_string .= "$primaryTag features ";
      }
      else{
        $print_feature_string .= "$tag feature ";
      }
   }
}

# Manage attributes if given
### If attributes given, parse them:
my @attListOk;
if ($opt_attribute){
  my @attList = split(/,/, $opt_attribute); # split at comma as separated value

  foreach my $attribute (@attList){
      push @attListOk, $attribute;
      print "$attribute attribute will be processed.\n";

  }
  print "\n";
}

# start with some interesting information
my $stringPrint = strftime "%m/%d/%Y at %Hh%Mm%Ss", localtime;
$stringPrint .= "\nusage: $0 @copyARGV\n";
$stringPrint .= "We will discard $print_feature_string that have the attribute $opt_attribute.\n";

if ($opt_output){
  print $ostreamReport $stringPrint;
  print $stringPrint;
}
else{ print $stringPrint; }
                                                      #######################
                                                      #        MAIN         #
#                     >>>>>>>>>>>>>>>>>>>>>>>>>       #######################       <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
my $cases=0;
######################
### Parse GFF input #
my ($hash_omniscient, $hash_mRNAGeneLink) =  slurp_gff3_file_JD({ input => $opt_gff,
                                                                  verbose => $opt_verbose
                                                                });
print("Parsing Finished\n");
### END Parse GFF input #
#########################

my $removeit=undef;
foreach my $tag_l1 ( sort keys %{$hash_omniscient->{'level1'}}){
  foreach my $id_l1 (sort keys %{$hash_omniscient->{'level1'}{$tag_l1} } ){

    my $feature_l1=$hash_omniscient->{'level1'}{$tag_l1}{$id_l1};

    $removeit = check_feature($feature_l1, 'level1', \@ptagList, \@attListOk, $opt_test);
    if ($removeit){
      remove_l1_and_subfeature($hash_omniscient, $feature_l1, $tag_l1, $id_l1);
      next;
    }

    #################
    # == LEVEL 2 == #
    #################
    foreach my $tag_l2 (sort keys %{$hash_omniscient->{'level2'}}){ # primary_tag_key_level2 = mrna or mirna or ncrna or trna etc...

      if ( exists_keys( $hash_omniscient, ('level2', $tag_l2, $id_l1) ) ){
        my @list_fl2 = @{$hash_omniscient->{'level2'}{$tag_l2}{$id_l1}};
        foreach my $feature_l2 ( @list_fl2 ) {

          $removeit = check_feature($feature_l2,'level2', \@ptagList, \@attListOk, $opt_test);
          if ($removeit){
            remove_l2_and_subfeature($hash_omniscient, $feature_l2, $tag_l1, $tag_l2, $id_l1);
            next;
          }
          #################
          # == LEVEL 3 == #
          #################
          my $id_l2 = lc($feature_l2->_tag_value('ID'));

          foreach my $tag_l3 (sort keys %{$hash_omniscient->{'level3'}}){ # primary_tag_key_level3 = cds or exon or start_codon or utr etc...
            if ( exists_keys( $hash_omniscient, ('level3', $tag_l3, $id_l2) ) ){
              my @list_fl3 = @{$hash_omniscient->{'level3'}{$tag_l3}{$id_l2}};
              foreach my $feature_l3 ( @list_fl3 ) {

                $removeit = check_feature($feature_l3, 'level3', \@ptagList, \@attListOk, $opt_test);
                if ($removeit){
                  remove_l3($hash_omniscient, $feature_l3, $tag_l1, $tag_l2, $tag_l3, $id_l1, $id_l2);
                }
              }
            }
          }
        }
      }
    }
  }
}

print_omniscient($hash_omniscient, $gffout_ok); #print gene modified


$stringPrint = "$cases features removed \n";
if ($opt_output){
  print $ostreamReport $stringPrint;
  print $stringPrint;
} else{ print $stringPrint; }

#######################################################################################################################
        ####################
         #     methods    #
          ################
           ##############
            ############
             ##########
              ########
               ######
                ####
                 ##

sub check_feature{
  my  ($feature, $level, $ptagList, $attListOk, $opt_test)=@_;

  my $removeit=undef;
  my $primary_tag=$feature->primary_tag;

  # check primary tag (feature type) to handle
  foreach my $ptag (@$ptagList){

    if($ptag eq "all"){
      $removeit = should_we_remove_feature($feature, $attListOk, $opt_test);
    }
    elsif(lc($ptag) eq $level){
      $removeit = should_we_remove_feature($feature, $attListOk, $opt_test);
    }
    elsif(lc($ptag) eq lc($primary_tag) ){
      $removeit = should_we_remove_feature($feature, $attListOk, $opt_test);
    }
  }
  return $removeit;
}

sub should_we_remove_feature{
  my ($feature, $attListOk, $opt_test)=@_;

  foreach my $att ( @{$attListOk} ){
    if ($feature->has_tag($att)){
      $opt_test ? return 0 : return 1;

    }
  }
  $opt_test ? return 1 : return 0;
}

# remove from omniscient l1 feature and all subfeatures
sub remove_l1_and_subfeature{
  my ($omniscient, $feature, $tag_l1, $id_l1)=@_;

  foreach my $ptag_l2 (keys %{$omniscient->{'level2'}}){ # primary_tag_key_level2 = mrna or mirna or ncrna or trna etc...

    if ( exists_keys( $omniscient, ('level2', $ptag_l2, $id_l1) ) ){
      foreach my $feature_l2 ( @{$omniscient->{'level2'}{$ptag_l2}{$id_l1}}) {

        my $level2_ID = lc($feature_l2->_tag_value('ID'));

        foreach my $ptag_l3 (keys %{$omniscient->{'level3'}}){ # primary_tag_key_level3 = cds or exon or start_codon or utr etc...
          if ( exists_keys( $omniscient, ('level3', $ptag_l3, $level2_ID) ) ){
            foreach my $feature_l3 ( @{$omniscient->{'level3'}{$ptag_l3}{$level2_ID}}) {
              $cases++; print $fhout_discarded $feature_l3->_tag_value('ID')."\n";
            }
            delete $omniscient->{'level3'}{$ptag_l3}{$level2_ID} # delete level3
          }
        }
        $cases++;  print $fhout_discarded $feature_l2->_tag_value('ID')."\n";
      }
      delete $omniscient->{'level2'}{$ptag_l2}{$id_l1} # delete level2
    }
  }
  $cases++; print $fhout_discarded $feature->gff_string()."\n";
  delete $omniscient->{'level1'}{$tag_l1}{$id_l1}; # delete level1
}

# remove from omniscient l2 feature and all subfeatures
sub remove_l2_and_subfeature{
  my ($omniscient, $feature, $ptag_l1, $ptag_l2, $id_l1)=@_;

  my $level2_Parent_ID = lc($feature->_tag_value('Parent'));
  my $level2_ID = lc($feature->_tag_value('ID'));

  if ( exists_keys( $omniscient, ('level2', $ptag_l2, $id_l1) ) ){ # just extra security in case
    foreach my $feature_l2 ( @{$omniscient->{'level2'}{$ptag_l2}{$id_l1}}) {

      if($level2_ID eq lc($feature_l2->_tag_value('ID')) ){
        # let's delete all l3 subfeatures before to remove the l2
        foreach my $ptag_l3 (keys %{$omniscient->{'level3'}}){ # primary_tag_key_level3 = cds or exon or start_codon or utr etc...
          if ( exists_keys( $omniscient, ('level3', $ptag_l3, $level2_ID)  ) ){
            foreach my $feature_l3 ( @{$omniscient->{'level3'}{$ptag_l3}{$level2_ID}}) {
              $cases++; print $fhout_discarded $feature_l3->_tag_value('ID')."\n";
            }
            delete $omniscient->{'level3'}{$ptag_l3}{$level2_ID} # delete level3
          }
        }
      }
    }

    # delete level2 and the hash pointer if the list is empty (no uisoform left)
    my @id_concern_list=($id_l1);
    my @id_list_to_remove=($level2_ID);
    my @list_tag_key=('all');
    $cases++; print $fhout_discarded $feature->gff_string()."\n";
    remove_element_from_omniscient(\@id_concern_list, \@id_list_to_remove, $omniscient, 'level2','false', \@list_tag_key);


    if( ! exists_keys($omniscient, ('level2', $ptag_l2, $id_l1) ) ){
      #The list was empty so l2 has been removed, we can now remove l1
      if( exists_keys($hash_omniscient, ('level1', $ptag_l1, $id_l1) ) ){
        $cases++; print $fhout_discarded $id_l1."\n";
        delete $hash_omniscient->{'level1'}{$ptag_l1}{$id_l1};
      }
    }
  }
}

# remove from omniscient l2 feature and all subfeatures
sub remove_l3{
  my ($omniscient, $feature, $ptag_l1, $ptag_l2, $ptag_l3, $id_l1, $id_l2)=@_;

  my $level3_Parent_ID = lc($feature->_tag_value('Parent'));
  my $id_l3 = lc($feature->_tag_value('ID'));

  if ( exists_keys( $omniscient, ('level3', $ptag_l3, $id_l2) ) ){ # just extra security in case
    foreach my $feature_l3 ( @{$omniscient->{'level3'}{$ptag_l3}{$id_l2}}) {

      if($id_l3 eq lc($feature_l3->_tag_value('ID')) ){
        #remove one feature and pointer if no more feature left in the list
        my @id_concern_list=($level3_Parent_ID);
        my @id_list_to_remove=($id_l3);
        my @list_tag_key=('all');
        $cases++; print $fhout_discarded $feature->gff_string()."\n";
        remove_element_from_omniscient(\@id_concern_list, \@id_list_to_remove, $omniscient, 'level3','false', \@list_tag_key);
      }
    }
  }

  # List empty check if we remove l2 or other l3 linked to it
  if( ! exists_keys($omniscient, ('level3', $ptag_l3, $id_l2)) ){
    foreach my $tag_l3 (keys %{$hash_omniscient->{'level3'}}){ # primary_tag_key_level3 = cds or exon or start_codon or utr etc...
      if ( exists_keys( $hash_omniscient, ('level3', $tag_l3, $id_l2) ) ){
        return;
      }
    }

    # if we arrive here it means no more L3 feature is attached to L2
    # we remove the L2 parent properly (if isoforms they are kept)
    my @id_concern_list=($id_l1);
    my @id_list_to_remove=($id_l2);
    my @list_tag_key=('all');
    $cases++; print $fhout_discarded $id_l2."\n";
    remove_element_from_omniscient(\@id_concern_list, \@id_list_to_remove, $omniscient, 'level2','false', \@list_tag_key);

    # Should we remove l1 too?
    if( ! exists_keys($omniscient, ('level2', $ptag_l2, $id_l1) ) ){
      #The list was empty so l2 has been removed, we can now remove l1
      if( exists_keys($hash_omniscient, ('level1', $ptag_l1, $id_l1) ) ){
        $cases++; print $fhout_discarded $id_l1."\n";
        delete $hash_omniscient->{'level1'}{$ptag_l1}{$id_l1}
      }
    }
  }
}

__END__

=head1 NAME

agat_sp_select_feature_by_attribute_presence.pl

=head1 DESCRIPTION

The script aims to filter features according to attribute presence (9th column).
If the attribute exists, the feature is discarded.
Attribute are stored in the 9th column and have this shape: tag=value
/!\ Removing a level1 or level2 feature will automatically remove all linked subfeatures, and
removing all children of a feature will automatically remove this feature too.

=head1 SYNOPSIS

    ./agat_sp_select_feature_by_attribute_presence.pl -f infile.gff -a <tag> [ --output outfile ]
    ./agat_sp_select_feature_by_attribute_presence.pl --help

=head1 OPTIONS

=over 8

=item B<-f>, B<--reffile>, B<--gff>  or B<-ref>

Input GFF3 file that will be read

=item B<-p>,  B<--type> or  B<-l>

primary tag option, case insensitive, list. Allow to specied the feature types that will be handled.
You can specified a specific feature by given its primary tag name (column 3) as: cds, Gene, MrNa
You can specify directly all the feature of a particular level:
      level2=mRNA,ncRNA,tRNA,etc
      level3=CDS,exon,UTR,etc
By default all feature are taking into account. fill the option by the value "all" will have the same behaviour.


=item B<--attribute>, B<--att>, B<-a>

String - Attributes tag specified will be used to filter the feature type (feature type can also be specified by the option -p). List of attribute tags must be coma separated.

=item B<--flip>

BOLEAN - In order to flip the test and keep features that do have the attribute and filter those without

=item B<-o> or B<--output>

Output GFF file.  If no output file is specified, the output will be
written to STDOUT.

=item B<-h> or B<--help>

Display this helpful text.

=back

=head1 FEEDBACK

=head2 Did you find a bug?

Do not hesitate to report bugs to help us keep track of the bugs and their
resolution. Please use the GitHub issue tracking system available at this
address:

            https://github.com/NBISweden/AGAT/issues

 Ensure that the bug was not already reported by searching under Issues.
 If you're unable to find an (open) issue addressing the problem, open a new one.
 Try as much as possible to include in the issue when relevant:
 - a clear description,
 - as much relevant information as possible,
 - the command used,
 - a data sample,
 - an explanation of the expected behaviour that is not occurring.

=head2 Do you want to contribute?

You are very welcome, visit this address for the Contributing guidelines:
https://github.com/NBISweden/AGAT/blob/master/CONTRIBUTING.md

=cut

AUTHOR - Jacques Dainat
